/*
 *	NXP PNX8550 watchdog driver. Based on SoftDog 0.07
 *
 *	(c) Copyright 1996 Alan Cox <alan@redhat.com>, All Rights Reserved.
 *				http://www.redhat.com
 *
 *
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Neither Alan Cox nor CymruNet Ltd. admit liability nor provide
 *	warranty for any of this software. This material is provided
 *	"AS-IS" and at no charge.
 *
 *	(c) Copyright 1995    Alan Cox <alan@lxorguk.ukuu.org.uk>
 *
 *	Software only watchdog driver. Unlike its big brother the WDT501P
 *	driver this won't always recover a failed machine.
 *
 *  03/96: Angelo Haritsis <ah@doc.ic.ac.uk> :
 *	Modularised.
 *	Added soft_margin; use upon insmod to change the timer delay.
 *	NB: uses same minor as wdt (WATCHDOG_MINOR); we could use separate
 *	    minors.
 *
 *  19980911 Alan Cox
 *	Made SMP safe for 2.3.x
 *
 *  20011127 Joel Becker (jlbec@evilplan.org>
 *	Added soft_noboot; Allows testing the softdog trigger without
 *	requiring a recompile.
 *	Added WDIOC_GETTIMEOUT and WDIOC_SETTIMOUT.
 *
 *  20020530 Joel Becker <joel.becker@oracle.com>
 *  	Added Matt Domsch's nowayout module option.
 *
 *  20070713 Mike Crowe <mcrowe@rokulabs.com>
 *      Port softdog to run on PNX8550 with the hardware watchdog watching it.
 */

/*
 * The PNX8550 watchdog isn't really that compatible with the Linux
 * watchdog API out of the box because depending on the CPU clock rate
 * its maximum period before expiring can be much less than a
 * minute. So rather than use the watchdog directly we work rather
 * like the software watchdog but using the software watchdog to
 * generate the equivalent of timer interrupts. The hardware watchdog
 * is used to force a reboot if the kernel fails to stop sending us
 * interrupts.
 *
 * One complication is that once we start the watchdog timer we can't
 * stop it so in order to comply with the Linux interface we have to
 * keep receiving interrupts and prodding the watchdog even if the
 * watchdog is disabled. This also means that if we're compiled as a
 * module we cannot be removed.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/fs.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/jiffies.h>
#include <linux/interrupt.h>

#include <asm/mach-pnx8550/glb.h>
#include <asm/mach-pnx8550/int.h>
#include <asm/uaccess.h>
#include <asm/mipsregs.h>

#define PFX "PNX8550dog: "

#define TIMER_MARGIN	60		/* Default is 60 seconds */
static int soft_margin = TIMER_MARGIN;	/* in seconds */
module_param(soft_margin, int, 0);
MODULE_PARM_DESC(soft_margin, "Watchdog pnx8550wdt_margin in seconds. (0<margin<65536, default=" __MODULE_STRING(TIMER_MARGIN) ")");

static int nowayout = WATCHDOG_NOWAYOUT;
module_param(nowayout, int, 0);
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");

#ifdef ONLY_TESTING
static int soft_noboot = 1;
#defined dprint(...) printk(__VA_ARGS__)
#else
static int soft_noboot = 0;
#define dprint(...) do {} while(0)
#endif  /* ONLY_TESTING */

module_param(soft_noboot, int, 0);
MODULE_PARM_DESC(soft_noboot, "Softdog action, set to 1 to ignore reboots, 0 to reboot (default depends on ONLY_TESTING)");

static int pnx8550wdt_stop(void);

/*
 *	Our timer
 */

volatile unsigned long last_keepalive; /* When we last saw a keepalive. Should really be atomic_t. */
static unsigned long driver_open;
static char expect_close;
static const unsigned long HARDWARE_COMPARE_VALUE = 0x7fffffff;

static irqreturn_t pnx8550wdt_interrupt(int cpl, void *private)
{
	dprint("WDT interrupt\n");

	// Acknowledge the interrupt by resetting the counter
	write_c0_compare3(HARDWARE_COMPARE_VALUE);

	// Timer 3 cannot be disabled so we have to deal with it even
	// if the interrupt happens when we're disabled.
	if (PNX8550_RST_EN_WATCHDOG != 0) {
		// If we have expired and we're open then we've fired. Even in
		// the face of wraparound this is safe.
		if (unlikely((jiffies - last_keepalive) > (soft_margin*HZ))) {
			if (soft_noboot) {
				printk(KERN_CRIT PFX "Triggered - Reboot ignored.\n");
			} else {
				printk(KERN_CRIT PFX "Initiating system reboot (jiffies=%lu, last_seen=%lu, diff=%lu).\n", jiffies, last_keepalive, jiffies - last_keepalive);
				emergency_restart();
				printk(KERN_CRIT PFX "Reboot didn't ?????\n");
			}
		}
	} else {
		dprint("Timer interrupt when disabled. Ignoring.\n");
	}	
        return IRQ_HANDLED;
}

/*
 *	Softdog operations
 */

static int pnx8550wdt_keepalive(void)
{
	last_keepalive = jiffies;
	return 0;
}

static int pnx8550wdt_start(void)
{
	dprint("Starting watchdog\n");
	PNX8550_RST_EN_WATCHDOG = 1;

	// This awakens counter 3.
	write_c0_compare3(HARDWARE_COMPARE_VALUE);
	write_c0_count3(0);
	
	pnx8550wdt_keepalive();
	return 0;
}

static int pnx8550wdt_stop(void)
{
	dprint("Stop watchdog\n");

	// There is no point in trying to stop the timer because once
	// we've started it then it can't be stopped. We'll just use
	// this as a flag for our own benefit.
	PNX8550_RST_EN_WATCHDOG = 0;
	return 0;
}

static int pnx8550wdt_set_heartbeat(int t)
{
	if ((t < 0x0001) || (t > 0xFFFF))
		return -EINVAL;

	soft_margin = t;
	return 0;
}

/*
 *	/dev/watchdog handling
 */

static int pnx8550wdt_open(struct inode *inode, struct file *file)
{
	if (test_and_set_bit(0, &driver_open))
		return -EBUSY;

	/*
	 *	Activate timer
	 */
	pnx8550wdt_start();
	return nonseekable_open(inode, file);
}

static int pnx8550wdt_release(struct inode *inode, struct file *file)
{
	/*
	 *	Shut off the timer.
	 * 	Lock it in if it's a module and we set nowayout
	 */
	if (expect_close == 42) {
		pnx8550wdt_stop();
		module_put(THIS_MODULE);
	} else {
		printk(KERN_CRIT PFX "Unexpected close, not stopping watchdog!\n");
		pnx8550wdt_keepalive();
	}
	clear_bit(0, &driver_open);
	expect_close = 0;
	return 0;
}

static ssize_t pnx8550wdt_write(struct file *file, const char __user *data, size_t len, loff_t *ppos)
{
	/*
	 *	Refresh the timer.
	 */
	if(len) {
		if (!nowayout) {
			size_t i;

			/* In case it was set long ago */
			expect_close = 0;

			for (i = 0; i != len; i++) {
				char c;

				if (get_user(c, data + i))
					return -EFAULT;
				if (c == 'V')
					expect_close = 42;
			}
		}
		pnx8550wdt_keepalive();
	}
	return len;
}

static int pnx8550wdt_ioctl(struct inode *inode, struct file *file,
	unsigned int cmd, unsigned long arg)
{
	void __user *argp = (void __user *)arg;
	int __user *p = argp;
	int new_margin;
	static struct watchdog_info ident = {
		.options =		WDIOF_SETTIMEOUT |
					WDIOF_KEEPALIVEPING |
					WDIOF_MAGICCLOSE,
		.firmware_version =	0,
		.identity =		"PNX8550 Watchdog",
	};
	switch (cmd) {
		default:
			return -ENOTTY;
		case WDIOC_GETSUPPORT:
			return copy_to_user(argp, &ident,
				sizeof(ident)) ? -EFAULT : 0;
		case WDIOC_GETSTATUS:
		case WDIOC_GETBOOTSTATUS:
			return put_user(0, p);
		case WDIOC_KEEPALIVE:
			pnx8550wdt_keepalive();
			return 0;
		case WDIOC_SETTIMEOUT:
			if (get_user(new_margin, p))
				return -EFAULT;
			if (pnx8550wdt_set_heartbeat(new_margin))
				return -EINVAL;
			pnx8550wdt_keepalive();
			/* Fall */
		case WDIOC_GETTIMEOUT:
			return put_user(soft_margin, p);
	}
}

/*
 *	Notifier for system down
 */

static int pnx8550wdt_notify_sys(struct notifier_block *this, unsigned long code,
	void *unused)
{
	if(code==SYS_DOWN || code==SYS_HALT) {
		/* Turn the WDT off */
		pnx8550wdt_stop();
	}
	return NOTIFY_DONE;
}

/*
 *	Kernel Interfaces
 */

static const struct file_operations pnx8550wdt_fops = {
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= pnx8550wdt_write,
	.ioctl		= pnx8550wdt_ioctl,
	.open		= pnx8550wdt_open,
	.release	= pnx8550wdt_release,
};

static struct miscdevice pnx8550wdt_miscdev = {
	.minor		= WATCHDOG_MINOR,
	.name		= "watchdog",
	.fops		= &pnx8550wdt_fops,
};

static struct notifier_block pnx8550wdt_notifier = {
	.notifier_call	= pnx8550wdt_notify_sys,
};

static char banner[] __initdata = KERN_INFO "PNX8550 Watchdog Timer: initialized. soft_noboot=%d soft_margin=%d sec (nowayout= %d)\n";

static int __init watchdog_init(void)
{
	int ret;

	/* Check that the soft_margin value is within it's range ; if not reset to the default */
	if (pnx8550wdt_set_heartbeat(soft_margin)) {
		pnx8550wdt_set_heartbeat(TIMER_MARGIN);
		printk(KERN_INFO PFX "soft_margin value must be 0<soft_margin<65536, using %d\n",
			TIMER_MARGIN);
	}

	ret = register_reboot_notifier(&pnx8550wdt_notifier);
	if (ret) {
		printk (KERN_ERR PFX "cannot register reboot notifier (err=%d)\n",
			ret);
		return ret;
	}

	ret = misc_register(&pnx8550wdt_miscdev);
	if (ret) {
		printk (KERN_ERR PFX "cannot register miscdev on minor=%d (err=%d)\n",
			WATCHDOG_MINOR, ret);
		unregister_reboot_notifier(&pnx8550wdt_notifier);
		return ret;
	}

	if (request_irq(PNX8550_INT_TIMER3, pnx8550wdt_interrupt, 0,
			"PNX8550 Watchdog", &pnx8550wdt_miscdev)) {
		printk(KERN_ERR "Failed to get WDT IRQ\n");
		unregister_reboot_notifier(&pnx8550wdt_notifier);
		misc_deregister(&pnx8550wdt_miscdev);
		return -EIO;
	}

	// We can't be unloaded once we're loaded because we can't
	// stop the timer.
	__module_get(THIS_MODULE);
	
	printk(banner, soft_noboot, soft_margin, nowayout);

	return 0;
}

static void __exit watchdog_exit(void)
{
	free_irq(PNX8550_INT_TIMER3, &pnx8550wdt_miscdev);
	misc_deregister(&pnx8550wdt_miscdev);
	unregister_reboot_notifier(&pnx8550wdt_notifier);
}

module_init(watchdog_init);
module_exit(watchdog_exit);

MODULE_AUTHOR("Mike Crowe");
MODULE_DESCRIPTION("PNX8550 Watchdog Device Driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);
