//@+leo-ver=4
//@+node:@file host/helpers.c
//@@language c
//@<< lud-base #includes >>
//@+node:<< lud-base #includes >>
#ifdef CONFIG_SMP
#define __SMP__
#endif

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/page.h>
#include <asm/semaphore.h>
#include <asm/atomic.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>

#include <linux/module.h>

#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/kmod.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#include <linux/device.h>
#include <linux/cdev.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18))
#include <linux/compiler.h>
#else
#define likely(x) (x)
#define unlikely(x) (x)
#endif /* LINUX_VERSION_CODE */

#include <stdarg.h>


#include "compat.h"
#include "lud_mmio.h"
#include "luddite-host.h"
#include "luddite/version.h"
#include "luddite/os.h"
#include "exported.h"

#include "core.h"
#include "procfs.h"
#include "node.h"
#include "helpers.h"
#include "channels.h"
//@-node:<< lud-base #includes >>
//@nl
//@<< lud-helpers declarations >>
//@+node:<< lud-helpers declarations >>
#ifdef __sparc__
/**
 * @internal
 * A Structure holding all active addresses/references
 */
static struct LUD_ADDRS {
	void *addr ;
} lud_addrs[MAX_LUD_ADDRS] ;

#endif

//@-node:<< lud-helpers declarations >>
//@nl
//@+others
//@+node:copy_to_user_fromio
//@+at
//@nonl
// * * Copy from Target memory (in PCI space) to userspace using page size
// bounce buffer * If successul, returns 0, else number of bytes left to copy
// (same as copy_to_user)
//@-at
//@@c

int copy_to_user_fromio(void *dst, void *src, register int len)
{
	register char *d = dst ;
	register char *s = src ;
	unsigned char *buffer ;

	if (!access_ok(VERIFY_WRITE, dst, len)) {
		return len ;
	}

#ifdef NO_BOUNCE_BUFFER
	if (unlikely(copy_to_user(dst, src, len)))  {
		return len;
	}
	return 0;
#else
	buffer = kmalloc(PAGE_SIZE,GFP_KERNEL);
	if (unlikely(!buffer))
		return len ;

	while (len > 0) {
		register int residual = (len > PAGE_SIZE) ? PAGE_SIZE : len ;

		fast_memcpy_fromio(buffer, s, residual);

		if (unlikely(copy_to_user(d, buffer, residual)))
			break ;		/* unlikely page fault */

		s += residual ;
		d += residual ;
		len -= residual ;
	}
	kfree(buffer);
	return len ;
#endif
}
//@-node:copy_to_user_fromio
//@+node:copy_from_user_toio


/**
 * Copy from userspace to Target memory (over PCI) using page size bounce buffer.
 * If successful, returns 0, else number of bytes left to copy (same as copy_from_user)
 */
int copy_from_user_toio(void *dst, void *src, register int len)
{
	register char *d = dst ;
	register char *s = src ;
	unsigned char *buffer ;

	if (!access_ok(VERIFY_READ, src, len)) {
		return len;
	}

#ifdef NO_BOUNCE_BUFFER
	if (unlikely(copy_from_user(dst, src, len))) {
		return len;
	}
	return 0;
#else
	buffer = kmalloc(PAGE_SIZE,GFP_KERNEL);
	if (unlikely(!buffer))
		return len;

	while (len > 0) {
		register int residual = (len > PAGE_SIZE) ? PAGE_SIZE : len ;

		if (unlikely(copy_from_user(buffer, s, residual)))
			break; 		/* unlikely page fault */

		fast_memcpy_toio(d, buffer, residual);

		s += residual ;
		d += residual ;
		len -= residual ;
	}
	kfree(buffer);
	return len;
#endif
}
//@-node:copy_from_user_toio
//@+node:strncpy_fromio
//@+at
//@nonl
// * * Strncpy from io space to kernel space. readb() can compile down to
// single * memory reads on most sane PCI architectures
//@-at
//@@c

int strncpy_fromio(void *dest, const void *src, register int n)
{
	register const char *s = src ;
	register char *d = dest ;
	register char ch ;

	while (((ch = readb((IOTYPE)s)) !=0) && (n>0)) {
		*d++ = ch ;
		s++;
		n--;
	}
	*d = '\0' ;

	return s - (char *)src ;

}
//@-node:strncpy_fromio
//@+node:strcpy_toio
//@+at
//@nonl
// * * Strcpy from kernel to io space
//@-at
//@@c

int strcpy_toio(void *dest, const void *src)
{
	const char *s = src ;
	char *d = dest ;
	char ch ;

	while ((ch = *s++) != 0) {
		writeb(ch, (IOTYPE)d);
		d++;
	}
	writeb('\0', (IOTYPE)d);

	return s - (char *)src ;
}
//@-node:strcpy_toio
//@+node:addr_to_token
//@+at
//@nonl
// * Convert pointers to 32 bit tokens - for 64 bit kernels such as SPARC
//@-at
//@@c

#ifdef __sparc__

#define LUD_ADDR_TOKEN_MAGIC	0x1234c0de

static spinlock_t  addr_lock;
int addr_to_token(void *addr)
{
	register int i ;
	register int token = -1 ;
	unsigned int irqstate;

	/* For large MAX_LUD_ADDRS, the spin lock should be replaced by a semaphore */

	spin_lock_irqsave(&addr_lock, irqstate);
	for (i=0;i<MAX_LUD_ADDRS;i++) {
		if (lud_addrs[i].addr == 0) {
			lud_addrs[i].addr = addr ;
			spin_unlock_irqrestore(&addr_lock,irqstate);
			return (i + LUD_ADDR_TOKEN_MAGIC) ;
		}
	}
	spin_unlock_irqrestore(&addr_lock,irqstate);

	printk(KERN_CRIT "Out of tokens (max %d)\n",MAX_LUD_ADDRS);

	return token ;
}
//@-node:addr_to_token
//@+node:token_to_addr
//@+at
//@nonl
// * Convert tokens back to pointers. if free == 1, free the token
//@-at
//@@c

void* token_to_addr(int token)
{
	void* addr = NULL ;

	token -= LUD_ADDR_TOKEN_MAGIC ;

	if (token >= 0 && token < MAX_LUD_ADDRS) {
		addr = lud_addrs[token].addr ;
	}

	return addr ;
}
//@-node:token_to_addr
//@+node:token_free
//@+at
//@nonl
// * Free token
//@-at
//@@c

void token_free(int token)
{
	unsigned int irqstate;

	token -= LUD_ADDR_TOKEN_MAGIC ;
	if (token >= 0 && token < MAX_LUD_ADDRS) {
		spin_lock_irqsave(&addr_lock, irqstate);
		lud_addrs[token].addr = NULL ;
		spin_unlock_irqrestore(&addr_lock,irqstate);
	}
}
#undef LUD_ADDR_TOKEN_MAGIC

EXPORT_SYMBOL(addr_to_token);
EXPORT_SYMBOL(token_to_addr);
EXPORT_SYMBOL(token_free);

#endif
//@nonl
//@-node:token_free
//@-others
//@-node:@file host/helpers.c
//@-leo
