/*
 * os abstraction layer/module implementation for luddite
 * based on Immo Benjes' os_abstraction stuff.
 */
#ifndef LUD_OS_DEF
#define LUD_OS_DEF

#include <luddite-host.h>  /*<tmhostdev.h> for LUD_REPLY */

/*
 * Options:
 * LUD_MEMORY_TRACKING_ALIASES: Use the tracking API for all allocations.
 *                              Define when compiling the module(s) you want
 *                              to track, or always to track all allocations.
 * LUD_MEMORY_TRACKING_REAL:    Define when compiling os.c to really track
 *                              mallocs - without this, the API is a no-op.
 *
 * These #defines are intended to allow flexibility:
 * - The no-op API allows binaries compiled with memory tracking to run
 *   on a kernel without memory tracking.
 * - Modules compiled without memory tracking can run even if the kernel
 *   has memory tracking.
 * - Memory tracking can be enabled on a per-source-file basis by enabling
 *   LUD_MEMORY_TRACKING_ALIASES from the Makefile where desired.
 */
#define LUD_MEMORY_TRACKING_ALIASES 
#define LUD_MEMORY_TRACKING_REAL 

#ifdef LUD_MEMORY_TRACKING_REAL
void *__lud_mem_tracking_set_and_check(const char *file,unsigned int line,
                                       void *userpointer) ;
void *__lud_mem_tracking_set(const char *file,unsigned int line,
                                       void *userpointer) ;

#  define lud_mem_tracking_set(file,line,userpointer)           \
          __lud_mem_tracking_set(file,line,userpointer)
#  define lud_mem_tracking_set_and_check(file,line,userpointer) \
          __lud_mem_tracking_set_and_check(file,line,userpointer)
#else
#  define lud_mem_tracking_set(file,line,userpointer) (userpointer)
#  define lud_mem_tracking_set_and_check(file,line,userpointer) (userpointer)
#endif

// aaded by k1f1
#define LUD_DEBUG_LEVEL 1

/*
 * Must disable LUD_MEMORY_TRACKING_ALIASES when compiling the
 * file luddite/kernel/target/lib/os.c
 *
 */
#ifdef LUD_COMPILING_OS_FILE
#ifdef LUD_MEMORY_TRACKING_ALIASES
#undef LUD_MEMORY_TRACKING_ALIASES
#endif
#endif

/**
 * @file
 */
/**
 * @defgroup group_os OS Abstraction Layer
 * @section os_intro Introduction
 * Luddite supplies a minimal OS abstraction layer, tailored to the
 * needs of the Luddite system. Instead of trying to implement the traditional
 * generic layer, we have chosen to implement a Luddite specific layer, ie.
 * a lot of the basic calls require arguments that are only relevent in the
 * Luddite context. This particular implementation of the layer is built
 * on top of pSOS.
 *
 * @section os_task Luddite Tasks
 * The Abstraction layer provides a series of functions for
 * starting and manipulating tasks. Tasks have names and a priority.
 * Tasks are identified by the LUD_TASK structure, which is used
 * as a parameter for all task functions. In general a NULL value used
 * as a task pointer results in the function being applied to the calling
 * task. 
 * Luddite tasks keep track of the system resources they are using,
 * and the luddite system has facilities for reporting this back to the host.
 *
 * @section os_prims Luddite OS Primitives
 * The Luddite OS abstraction layer supplies two general OS primitives.
 *  - LUD_SEMAPHORE is the luddite version of the semaphore primitive.
 *    The semaphore can be given an initial value, supports the signal and
 *    wait operations, where the wait operation can cause the calling task to
 *    block. The system keeps track of the number of semaphores created by
 *    any module. Being blocked on a semaphore is a known task state, and is
 *    available to the host, as an aid to debugging.
 *  - LUD_QUEUE is a simple queue primitive.
 *    Any data type can be appended to the end of a queue. Reading from an
 *    empty queue can cause a calling task to block, but it is possible
 *    to specify a timeout. Like semaphores, the number of queues created by
 *    a module is logged, and being blocked waiting on a queue is recognised
 *    task state.
 *
 * @section os_mem Memory Management
 * Memory allocation is a very important part of the OS abstraction layer.
 * Luddite offers functions for allocating areas of memory, with
 * the added benefits of specifying alignment requirements, and logging
 * which module is doing the allocating. The luddite system allows the
 * host to query memory usage, and will also report if a module is unloaded
 * without freeing all the memory it has allocated.
 *
 * @section os_time Time
 * Luddite now has one concept of current time, returned using the
 * lud_time() function call. This returns the number of 90KHz ticks
 * since startup as two 32 bit integers. This is a signed number, so
 * we only have 63 bits available. This means luddite time will wrap
 * in just over 3 million years. 
 * 
 * @{
 */

/**
 * @internal
 * The size of the data transfer area used for the /proc interface
 */
#define PROC_BUFFER_SIZE 4096

/**
 * Used as a generic function pointer type.
 */
typedef void (*LUD_FUNCTION)(void);
typedef struct LUD_TASK_S	LUD_TASK;
typedef struct LUD_MODULE_S	LUD_MODULE;

/**
 */

/**
 * A Luddite semaphore structure.
 */
typedef struct LUD_SEMAPHORE_S {
	LUD_MODULE*	mod; /**< the module that created the semaphore */
	unsigned long	sem; /**< pSOS semaphore */
} LUD_SEMAPHORE;

/**
 * A Luddite queue structure.
 */
typedef struct LUD_QUEUE_S {
	LUD_MODULE*	mod; /**< the module that created the queue */
	unsigned int	q; /**< the pSOS queue identifier */
} LUD_QUEUE;

/**
 * This is the maximum allowed length of the name string, passed to the
 * task creation function.
 */
#define LUD_TASK_NAME_LENGTH	16

/**
 * The possible states of a Luddite task
 */
enum TASK_STAT {
   CREATED, /**< task created but not yet running. */
   RUNNING, /**< task running. */
   PAUSED, /**< task paused by a system call (sleeping). */
   BLOCKED_SEM, /**< task blocked, waiting on a semaphore. */
   BLOCKED_QUEUE, /**< task blocked, waiting on a queue. */
   BLOCKED_POOL, /**< task blocked waiting for a buffer from a buffer pool. */
   NOT_CALLED /**< task very idle. */
};

#define LUD_TASK_MAX_REGISTERS	16

/**
 * The main task structure.
 * Contains all relevent information about a task, as well as the
 * tasks resource utilisation.
 */

struct LUD_TASK_S {
	LUD_MODULE*	module; /**< pointer to the owner of the task */ 
	LUD_TASK*	parent; /**< pointer to the task that created this one */
	LUD_QUEUE*	reply; /**< queue used to wait for replies from others */
	LUD_TASK*	children; /**< start of a list of children, created by this task */
	LUD_TASK*	next; /**< next child in a list */

	int		index;
	char		name[LUD_TASK_NAME_LENGTH]; /**< the tasks name */

	unsigned long	tid; /**< pSOS task identifier */
	unsigned long	cycles; /**< place to keep track of task CPU usage */

	short		status; /**< current run status */
	unsigned long	priority; /**< current pSOS priority */

	unsigned long	switches; /**< number of context switches to this task */

	void*		pools;	/**< keep track of buffer pools for this task */

	unsigned long	registers[LUD_TASK_MAX_REGISTERS] ;	/**< registers for this task */
};

#ifndef LUD_PORT_AND_NODE_TYPEDEF_DONE
#define LUD_PORT_AND_NODE_TYPEDEF_DONE
typedef struct LN_PORT LUD_PORT;
typedef struct LN_ENTRY LUD_NODE;
#endif

#define LUD_MODULE_EMBEDDED	0x08000
#define LUD_MODULE_STATIC    	0x10000

/**
 * The main luddite module structure
 * Used to track resources used by modules
 */
struct LUD_MODULE_S {
	int			major; /**< identifier used by the host */
	LUD_PORT*		port;  /**< port to receive commands from host */
	LUD_NODE*		node; /**< node used to target commands from host */
	char*           	name; /**< module name - as sent by host */
	void            	(*func)(void); /**< the modules  init_function */
	char*           	optstring; /**< options passed from host */
	int             	flags; 
	LUD_TASK*		main; /**< task used to run init_function */

	int			num_queues; /**< number of queues used */
	int			num_semaphores; /**< number of semaphores used */

	unsigned int		sem; /**< mutex sem for locking updates */
	unsigned int*		memory; /**< link to memory allocated by module */
	struct lud_prop_* 	props; /**< properties defined by this module */ 
	void*			params; /**< pointer to module parameters */

	int			clean_fail; /**< flag if a module clean failed */
	LUD_REPLY		load_reply; /**< reply state when loading module */
	LUD_REPLY		unload_reply; /**< reply state when unloading module */
};

/**
 * The maximum number of modules running at any one time.
 */
#define LUD_MAX_MODULES		16

/**
 * @returns The amount of memory (in bytes) currently used by the given module
 * @param m 		A pointer to the module we wish to query.
 */
extern unsigned int     lud_module_mem_usage(LUD_MODULE* m);

/**
 * Internal function used to dump module resource data into a buffer.
 */
extern void		lud_module_print_usage_info(char* buf);
/**
 * Internal function used to dump task data into a buffer.
 */
extern void		lud_module_print_task_info(char* buffer);

/**
 * Internal function used to dump buffer pool data into a buffer.
 */
extern void		lud_module_print_pool_info(char* buffer);

/**
 * @internal
 * Find a module based on a given name. The name be of the shortened form
 * accepted by @ref l_insmod and @ref l_rmmod commands.
 * @param name 		Name of the module
 * @returns 		A pointer to a LUD_MODULE structure or NULL 
 *			if the module cannot be found.
 */
extern LUD_MODULE*	lud_module_by_name(const char * name);

/**
 * @internal
 * Array used to store all the LUD_MODULE structures
 */
extern LUD_MODULE	TmModules[LUD_MAX_MODULES];

/**
 * @internal
 * Array used to map IRQ names to IRQ numbers
 */
extern char *		lud_irqmap[];

/**
 * @internal
 * Initialise luddite task structures
 */
extern void		lud_task_init(void);

/**
 * @internal
 * Add a buffer pool to the task structure.
 * Internal function used by all buffer_pool_create functions.
 */
extern void		lud_task_pool_add(LUD_TASK*,void*);

/**
 * @internal
 * Remove a buffer pool from a task structure.
 * Internal function used by all buffer_pool_delete functions.
 */
extern void		lud_task_pool_remove(LUD_TASK*,void*);

/**
 * Get a pointer to the LUD_MODULE structure of the module that started the
 * calling task.
 * @returns 		A pointer to a LUD_MODULE structure.
 */
extern LUD_MODULE*      lud_task_module_info(void);

/**
 * @returns a pointer to the callers LUD_TASK structure.
 */
extern LUD_TASK*        lud_task_task_info(void);

/**
 * @internal
 * Internal function used to create a special task to create a modules.
 * Semaphore protected - do not call from interrupt context.
 * @param major   	The module number.
 * @param name    	The task name.
 * @param f       	The main function for the task.
 * @param pri     	The priority for the task.
 * @param opts    	The module options.
 * @return        	The new task.
 * @todo FIXME error handling.
 */
extern LUD_TASK*        lud_task_create_module_task(int,char*,LUD_FUNCTION f,int pri,char* opt);

/**
 * Create and start a new task. Semaphore protected - do not call 
 * from interrupt context.
 * @param name 		The name of the task (used for info purposes only).
 * @param f 		The function to call as the task.
 * @param pri 		The default priority of the task.
 *
 * This function also excepts (up to four) additional parameters, that will
 * be passed to the given function when it is started.
 * @returns 		A pointer to a LUD_TASK structure, 
 *			representing the created task.
 */
extern LUD_TASK*        lud_task_create(char* name,LUD_FUNCTION f,int pri,...);

/**
 * @internal
 * Get a register value of a given task. 
 * @param tid		The id of the task, NULL means the current task.
 * @param reg		The register to get the value of (0..15).
 * @param ret		Pointer to a long to receive the value of the register.
 * @returns		0 OK, -1 failed (bad task id).
 */
extern int 		lud_task_getreg(unsigned long tid, int reg, unsigned long *ret);

/**
 * @internal
 * Set a register value of a given task
 * @param tid		The id of the task, NULL means the current task.
 * @param reg		The register to get the value of (0..15).
 * @param val		The value to be set to the register.
 * @returns		0 OK, -1 failed (bad task id).
 */
extern int 		lud_task_setreg(unsigned long tid, int reg, unsigned long val);

/**
 * Kill and delete the given task.
 * @param task 		Pointer to the task to be stopped. If NULL then 
 *			the caller is killed.
 * @returns 		0 if succeeded, else OS error number.
 */
extern int              lud_task_delete(LUD_TASK* task);

/**
 * Get the current priority of the given task
 * @param task 		Pointer to LUD_TASK structure of the given task 
 * 			If NULL then the caller task's priority is returned.
 * @returns int 	Priority level of the task.
 */
extern int              lud_task_get_priority(LUD_TASK* task);

/**
 * Set the priority of the given task.
 * @param task 		Pointer to LUD_TASK structure of the given task.
 * 			If NULL then the caller task's priority is set.
 * @param priority	Priority to be set. 
 * @returns 		0 if succeeded, else OS error number.
 */
extern int              lud_task_set_priority(LUD_TASK* task,int priority);

/**
 * Suspend the given task. Execution can be restarted using lud_task_resume().
 * @param task 		Pointer to LUD_TASK structure of the given task 
 *			to be suspended. If NULL then the caller is suspended.
 * @returns 		0 if succeeded, else OS error number.
 */
extern int		lud_task_suspend(LUD_TASK* task);

/**
 * Continue execution of a task stopped by lud_task_suspend().
 * @param task 		Pointer to LUD_TASK structure of the given task.
 *			to be resumed. If NULL then the caller is resumed.
 * @returns 		0 if succeeded, else OS error number.
 */
extern int		lud_task_resume(LUD_TASK* task);

/**
 * Suspend execution of the caller for the given time.
 * @param t the time to stop (1/100ths of a second).
 */
extern void		lud_task_sleep(int t);

/**
 * Get the named task id.
 * @param name 		Name of task to find. If NULL, the task id of the 
 *			calling task is returned.
 * @param tid		Address of int to receive returned task id.
 * @return 		0 if succeded, else 1 if failed.
 *
 */
extern int		lud_task_id(char *name, int *tid);

/**
 * @returns 		The CPU cycle counter (divided by 256)
 */
extern unsigned long 	lud_target_cycles(void);

/**
 * Create a semaphore. Semaphore protected - do not call from interrupt context.
 * @param inival 	Ihe initial value of the semaphore.
 * @returns 		A LUD_SEMAPHORE pointer to the created semaphore 
 *			(NULL on error).
 */
extern LUD_SEMAPHORE*   lud_sem_create(int inival);

/**
 * Memory allocation tracking version of lud_sem_create().
 *
 * Semaphore protected - do not call from interrupt context.
 *
 * @param file 		The __FILE__ value from the caller.
 * @param line 		The __LINE__ value from the caller.
 * @param inival 	Ihe initial value of the semaphore.
 * @return 		New semaphore, or NULL on error.
 */
extern LUD_SEMAPHORE*   lud_sem_tracking_create(const char *file,unsigned int line,int inival);
#ifdef LUD_MEMORY_TRACKING_ALIASES
#define lud_sem_create(inival) lud_sem_tracking_create(__FILE__,__LINE__,inival)
#endif

/**
 * Delete a semaphore.  Note that after this call, the semaphore has been
 * deleted and the #LUD_SEMAPHORE pointer passed as a paramater is invalid.
 * The module author <b>must</b> ensure that the #LUD_SEMAPHORE pointer is
 * never used again in any Luddite call.
 *
 * Semaphore protected - do not call from interrupt context.
 *
 * @param sem 		Pointer to the LUD_SEMAPHORE to delete.
 */
extern void             lud_sem_delete(LUD_SEMAPHORE* sem);

/**
 * Semaphore wait operation. Do not call from interrupt context.
 * @param sem 		Pointer to the LUD_SEMAPHORE to wait on
 * If the semaphore's current value is positive, it will be decremented.
 * Otherwise the caller will block until the operation can be carried out
 * (i.e. someone else calls lud_sem_signal().) 
 */
extern void             lud_sem_wait(LUD_SEMAPHORE* sem);

/**
 * Semaphore signal operation. Safe to call from interrupts.
 * @param sem 		Pointer to the LUD_SEMAPHORE to wait on
 * Increment the value of the given semaphore, will cause any blocked
 * tasks to try again if the value changes from 0 to 1
 */
extern void             lud_sem_signal(LUD_SEMAPHORE* sem);

/**
 * @internal
 * Initialise luddite queues
 */
extern void		lud_queue_init(void);

/**
 * Create a message queue.
 * Semaphore protected - do not call from interrupt context.
 *
 * @return 		The new queue, or NULL on error.
 */
extern LUD_QUEUE*       lud_queue_create(void);

/**
 * Memory allocation tracking version of lud_queue_create().
 * Semaphore protected - do not call from interrupt context.
 * @param file		Caller source file (from __FILE__) macro.
 * @param line		Caller line of source (from __LINE__) macro.
 * @return 		The new queue, or NULL on error.
 */
extern LUD_QUEUE*       lud_queue_tracking_create(const char *file,unsigned int line);
#ifdef LUD_MEMORY_TRACKING_ALIASES
#define lud_queue_create(inival) lud_queue_tracking_create(__FILE__,__LINE__)
#endif

/**
 * Delete a message queue. Note that after this call, the queue has been
 * deleted and the #LUD_QUEUE pointer passed as a paramater is invalid.
 * The module author <b>must</b> ensure that the #LUD_QUEUE pointer is
 * never used again in any Luddite call.
 *
 * Semaphore protected - do not call from interrupt context.
 *
 * @param q 		The queue to be deleted. Must not be NULL.
 */
extern void          	lud_queue_delete(LUD_QUEUE* q);

/**
 * Put one (int sized) item at the end of a queue.
 * The variable args are used to avoid type checking of queue data item.
 *
 * Safe to call from interrupt context.
 *
 * @param q 		The queue to send the data down.
 * @param Varargs 	A single integer or pointer to send.
 * @return 0 		on success, nonzero on error.
 */
extern int              lud_queue_send(LUD_QUEUE* q,...);

/**
 * Put one  (int sized) item onto the end of the urgent queue.
 * The variable args are used to avoid type checking of queue data item.
 *
 * Safe to call from interrupt context.
 *
 * This function puts the data onto the queue before any normal (non-
 * urgent queue items). The data is in inserted at the end of any previous
 * urgent queue items, before the first non-urgent queue item.
 * In contrast, lud_queue_send() puts the data at the end of the queue.
 *
 * Only use this function if you have a good reason for not
 * using lud_queue_send().
 *
 * @param q 		The queue to send the data down.
 * @param Varargs 	A single integer or pointer to send.
 * @return 		0 on success, nonzero on error.
 */
extern int              lud_queue_send_urgent(LUD_QUEUE* q,...);

/**
 * Get an entry from a queue.
 *
 * Safe to call from interrupt context iff you use LUD_NOWAIT. If timeout
 * is LUD_WAIT, the task will block until data is available on the queue. If
 * timeout is LUD_WAIT and no data is available the function returns
 * immediately with no data. If timeout is any other value, the function
 * will block until data is available (return is 1) or the timeout clock
 * ticks pass (return is zero).
 *
 * @param q 		The queue.
 * @param ret 		Destination for the recieved value.
 * @param timeout 	A timeout value in ticks:
 *  		           LUD_WAIT to wait forever, or
 *             		   LUD_NOWAIT to poll without blocking, or
 *			   or time to wait in clock ticks.
 * @return number of items received, negative number on error.
 *
 */
extern int              lud_queue_recv(LUD_QUEUE* q,unsigned int* ret,int timeout);

/**
 * Receive an urgent message (if it exists) from the head of a queue.
 *
 * @param q 		The queue
 * @param ret 		Where to store any received message.
 * @returns 		The number of items received (1 or 0).
*/
extern int              lud_queue_recv_urgent(LUD_QUEUE* q,unsigned int* ret);

/**
 * Returns the number of items on a queue.
 *
 * @param q 		The Luddite queue.
 * @return the number of items on the queue.
 * 
 */
extern int              lud_queue_length(LUD_QUEUE* q);

/**
 * @internal
 * Get queue statistics.
 *
 * @param buf   	Page sized buffer to write statics into.
 * @return      	Number of bytes written.
 */
int lud_queue_stat(char *buf);

/**
 * Wait for event(s) to be received. Waits for events to be sent in a
 * bit encoded 32 bit long. If timeout is zero, the task waits until
 * the specified event bits are received. If timeout is non-zero, the task
 * waits until timeout jiffies has passed, then returns with a non-zero
 * error return.
 *
 * @param event 	Messages bitmap.
 * @param timeout  	Timeout in clock jiffies. If zero, wait forever.
 * @param ev_rx    	Received events, bit-encoded.
 * @return 		0 is okay, else OS error code.
 */
int lud_event_receive(unsigned long event, unsigned long timeout, unsigned long *ev_rx);

/**
 * Send bit-encoded events to a task.   If a task is waiting on
 * event bits sent to it via lud_send_event the task is unblocked
 * and made runnable.  If a task is not waiting then the event bits
 * are logically or'd with the current set of pending events for that
 * task.
 *
 * @param tid  		Task id of the task to receive the event.
 * @param ev 		Event bit mask: use bits 0 to 15.
 */
void lud_event_send(int tid, unsigned long ev);

/**
 * Register an interrupt handler.
 *
 * @param irqname 	Interrupt name.
 * @param pri 		Interrupt Priority, 0-7, 0 lowest, 7 = highest (NMI).
 * @param level 	0 = Level Triggered, 1 = Edge Triggered.
 * @param isr 		The interrupt handler function.
 * @returns 		0 on success, anything else a failure.
 */
extern int lud_os_register_isr(char *irqname,int pri,int level,LUD_FUNCTION isr);

/**
 * Free a previously registered isr.
 *
 * @param irqname	Interrupt name.
 * @returns 		0 on success, anything else an error.
 */
extern int lud_os_unregister_isr(char *irqname);

/**
 * Allocate memory, setting memory to zero.
 * Semaphore protected - do not call from interrupt context.
 * @param size 		Number of bytes to allocate.
 * @returns 		Address of allocated memory or NULL if no 
 *			memory available.
 */
extern void*            lud_mem_alloc(unsigned int size);

/**
 * Memory tracking version of lud_mem_alloc().
 * Allocate memory, setting memory to zero.
 * Semaphore protected - do not call from interrupt context.
 * @param file		Caller source file (from __FILE__) macro .
 * @param line		Caller line of source (from __LINE__) macro.
 * @param size 		Number of bytes to allocate.
 * @returns 		Address of allocated memory or NULL if no 
 *			memory available.
 */
extern void*            lud_mem_tracking_alloc(const char *file,unsigned int line,unsigned int size);
#ifdef LUD_MEMORY_TRACKING_ALIASES
#define lud_mem_alloc(size)                   lud_mem_tracking_alloc(__FILE__,__LINE__,size)
#endif

/**
 * Allocate memory, but aligned to the cache row size of the TM
 * processor, also make sure the area allocated is a multiple of the cache
 * row size, for extra paranoia. Memory is set to zero. 
 * Semaphore protected - do not call from interrupt context.
 * @param 		size number of bytes to allocate, rounded up to 
 *			whole cache line size.
 * @returns 		address of allocated memory or NULL if no 
 *			memory available.
 */
extern void*            lud_mem_alloc_cache(unsigned int size);

/**
 * Memory tracking version of lud_mem_alloc_cache().
 * Allocate memory, but aligned to the cache row size of the TM
 * processor, also make sure the area allocated is a multiple of the cache
 * row size, for extra paranoia. Memory is set to zero. 
 * Semaphore protected - do not call from interrupt context.
 * @param file		Caller source file (from __FILE__) macro.
 * @param line		Caller line of source (from __LINE__) macro.
 * @param 		size number of bytes to allocate, rounded up to 
 *			whole cache line size.
 * @returns 		address of allocated memory or NULL if no 
 *			memory available.
 */
extern void*            lud_mem_tracking_alloc_cache  (const char *file,unsigned int line,unsigned int size);
#ifdef LUD_MEMORY_TRACKING_ALIASES
#define lud_mem_alloc_cache(size)             lud_mem_tracking_alloc_cache(__FILE__,__LINE__,size)
#endif

/**
 * Allocate memory, aligned to the given boundary. Unlike
 * lud_mem_alloc_cache(), the size of memory returned is only gauranteed to be
 * the size requested. Memory is set to zero.
 * Semaphore protected - do not call from interrupt context.
 * @param size 		Number of bytes to allocate.
 * @param alignment 	Alignment in bytes.
 * @returns 		Address of allocated memory or NULL if no memory available.
 */
extern void*            lud_mem_alloc_aligned(unsigned int size,unsigned int alignment);

/**
 * Memory tracking version of lud_mem_alloc_aligned().
 * Allocate memory, aligned to the given boundary. Unlike
 * lud_mem_alloc_cache(), the size of memory returned is only gauranteed to be
 * the size requested. Memory is set to zero.
 * Semaphore protected - do not call from interrupt context.
 * @param file		Caller source file (from __FILE__) macro .
 * @param line		Caller line of source (from __LINE__) macro.
 * @param size 		Number of bytes to allocate.
 * @param alignment 	Alignment in bytes.
 * @returns 		Address of allocated memory or NULL if no memory available.
 */
extern void*            lud_mem_tracking_alloc_aligned(const char *file,unsigned int line,unsigned int size,unsigned int alignment);
#ifdef LUD_MEMORY_TRACKING_ALIASES
#define lud_mem_alloc_aligned(size,alignment) lud_mem_tracking_alloc_aligned(__FILE__,__LINE__,size,alignment)
#endif

/**
 * Re-allocate size memory. 
 * Semaphore protected - do not call from interrupt context.  If size is 
 * zero and p != NULL, old memory block is freed. If block is expanded in
 * size, extra memory is set to 0.  Original data is copied to newly allocated
 * block.
 * @param p     	Old memory block, or NULL if none.
 * @param size  	Size of new memory block in bytes, or 0 if 
 *			none needed (NULL will be returned).
 * @return		Address of allocated memory, or NULL if out of 
 *			memory (p will not be affected), or NULL if size 
 *			is 0 (freeing p if p != NULL).
 */
extern void*            lud_mem_realloc(void* p,unsigned int size);

/**
 * Memory tracking version of lud_mem_realloc().
 * Re-allocate size memory. 
 * Semaphore protected - do not call from interrupt context.  If size is 
 * zero and p != NULL, old memory block is freed. If block is expanded in
 * size, extra memory is set to 0.  Original data is copied to newly allocated
 * block.
 * @param file		Caller source file (from __FILE__) macro.
 * @param line		Caller line of source (from __LINE__) macro.
 * @param p     	Old memory block, or NULL if none.
 * @param size  	Size of new memory block in bytes, or 0 if 
 *			none needed (NULL will be returned).
 * @return		Address of allocated memory, or NULL if out of 
 *			memory (p will not be affected), or NULL if size 
 *			is 0 (freeing p if p != NULL).
 */
extern void*            lud_mem_tracking_realloc      (const char *file,unsigned int line,void* p,unsigned int size);
#ifdef LUD_MEMORY_TRACKING_ALIASES
#define lud_mem_realloc(p,size)               lud_mem_tracking_realloc(__FILE__,__LINE__,p,size)
#endif

/**
 * Free memory allocated by one of the lud_mem_alloc* functions.
 * Semaphore protected - do not call from interrupt context.
 * @param ptr 		Memory to free, or NULL (no action).
 */
extern void             lud_mem_free(void* ptr);

/**
 * Check a region created by lud_mem_alloc* for buffer overruns, etc.
 * Checks if the memory was allocated through lud_mem_alloc* functions, and
 * sanity checks the guard bytes at the beginning of the memory block.
 * @param ptr		Block to check.
 * @param str		Message to use in debug output via an urgent lud_printk.
 * @returns 		Nonzero (true) if memory is OK, zero (false) if 
 *			memory is not a Luddite-allocated block or if it is 
 *			corrupted.
 */
extern int		lud_mem_check(void* ptr, char* str);

/**
 * @internal
 * Internal function to init the OS subsystem
 */
extern void		lud_os_init(void);

/**
 * @internal
 * Internal function to init some TCS stuff needed for loading modules.
 */
extern void		lud_os_fs_install(void);

/**
 * @internal
 * Internal function used to create a memory region for load ing a module.
 */
extern unsigned char*	lud_os_setup_load(int);

/**
 * @internal
 * Called if lud_assert() fails.
 * Prints a message and suspends this task.  Never returns.
 *
 * Should not be called directly from user code.  Call indirectly via the
 * lud_assert() macro.
 *
 * Uses lud_dprintk() - do not call from interrupt context.
 *
 * @param filename   The __FILE__ macro from the place the assertion failed.
 * @param linenumber The __LINE__ macro from the place the assertion failed.
 * @param condition  The text of the assertion which failed.
 */
extern void             lud_assert_failed(const char *filename, int linenumber, const char *condition);

#if LUD_DEBUG_LEVEL>1
/**
 * Check that a condition holds, and cause a run-time error if not.
 *
 * The run-time error includes the source code file and line number, as
 * well as the condition that failed.  On error, the current thread will
 * be suspended.
 *
 * Defining the LUD_DEBUG_LEVEL<2 will disable this check.  Do not
 * specify expressions with side-effects as the argument, since they will
 * not be evaluated if LUD_DEBUG_LEVEL<2.
 *
 * Uses printk() - do not call from interrupt context.
 *
 * @param _condition The condition which is expected to be true (nonzero).
 */
#define lud_assert(_condition) \
        do { \
                if (!(_condition)) { \
                        lud_assert_failed(__FILE__, __LINE__, #_condition);\
                } \
        } while (0)
#else
#define lud_assert(_condition)   do {} while (0)
#endif

/**
 * Check that a pointer is a valid pointer to the start of a memory block allocated
 * with lud_mem_alloc(), and cause a run-time error if not.
 *
 * The run-time error includes the source code file and line number, as
 * well as the name of the pointer that was invalid.  On error, the current thread will
 * be suspended.
 *
 * Defining the NDEBUG preprocessor symbol will disable this check.  Do not
 * specify expressions with side-effects as the argument, since they will
 * not be evaluated if NDEBUG is enabled.
 *
 * Uses printk() - do not call from interrupt context.
 *
 * @param _pointer 	The pointer which is expected to be valid and non-NULL.
 */
#define lud_mem_assert(_pointer) lud_assert(lud_mem_check(_pointer, #_pointer))

/**
 * @internal
 * Display all the memory allocated by the specified module.  Output is
 * via lud_printk().
 *
 * This is a no-op unless LUD_MEMORY_TRACKING_REAL is defined when os.c
 * is compiled.
 *
 * Warning: This is an O(N^2) algorithm. I can't figure out a safe O(N)
 * algorithm.
 *
 * @param m 		The module to query.
 */
void lud_mem_tracking_dump_module(LUD_MODULE* m);

/**
 * @internal
 * Display all the memory that is currently allocated.  Output is
 * via lud_printk().
 *
 * This is a no-op unless LUD_MEMORY_TRACKING_REAL is defined when os.c
 * is compiled.
 */
void lud_mem_tracking_dump_all(void);

/**
 * Returns the current luddite time
 * Luddite time is the number of 90KHz ticks since luddite started. It is stored
 * as a 64 bit value which takes over 3 million years to wrap.
 * @param hi 		Pointer to where the top 32 bits of the time 
 *			should be stored.
 * @param lo 		Pointer to where the bottom 32 bits of the time 
 *			should be stored.
 */
void lud_time(unsigned int* hi,unsigned int* lo);

/**
 * Sets the luddite clock. The values should be 90KHz ticks
 * Note the top bit of "hi" should always be zero.
 */

void lud_time_set(unsigned int hi,unsigned int lo);

/**
 * @}
 */

#endif /* LUD_OS_DEF */
